@@ -10,14 +10,17 @@ use std::{
     rc::Rc,
 };
 use swc_atoms::{js_word, JsWord};
-use swc_common::{collections::AHashMap, FileName, Mark, Span, DUMMY_SP};
+use swc_common::{
+    collections::{AHashMap, AHashSet},
+    FileName, Mark, Span, DUMMY_SP,
+};
 use swc_ecma_ast::*;
 use swc_ecma_transforms_base::helper;
 use swc_ecma_utils::{
     ident::IdentLike, member_expr, private_ident, quote_ident, quote_str, var::VarCollector,
     DestructuringFinder, ExprFactory, IsDirective,
 };
-use swc_ecma_visit::{noop_fold_type, Fold, FoldWith, VisitWith};
+use swc_ecma_visit::{noop_fold_type, noop_visit_type, Fold, FoldWith, Visit, VisitWith};
 
 pub fn common_js(
     top_level_mark: Mark,
@@ -55,6 +58,73 @@ where
     }
 }
 
+struct LazyIdentifierVisitor {
+    scope: Rc<RefCell<Scope>>,
+    top_level_idents: AHashSet<JsWord>,
+}
+impl LazyIdentifierVisitor {
+    fn new(scope: Rc<RefCell<Scope>>) -> Self {
+        LazyIdentifierVisitor {
+            scope,
+            top_level_idents: Default::default(),
+        }
+    }
+}
+
+/*
+An import can be performed lazily if it isn't used at the module's top-level.
+This visitor scans only the identifiers at the top level by not traversing nodes
+that introduce nesting. It then looks up identifiers to see if they match an
+imported specifier name.
+ */
+impl Visit for LazyIdentifierVisitor {
+    noop_visit_type!();
+
+    fn visit_import_decl(&mut self, _: &ImportDecl) {}
+    fn visit_export_decl(&mut self, _: &ExportDecl) {}
+    fn visit_named_export(&mut self, _: &NamedExport) {}
+    fn visit_export_default_decl(&mut self, _: &ExportDefaultDecl) {}
+    fn visit_export_default_expr(&mut self, _: &ExportDefaultExpr) {}
+
+    fn visit_export_all(&mut self, export: &ExportAll) {
+        self.top_level_idents.insert(export.src.value.clone());
+    }
+
+    fn visit_labeled_stmt(&mut self, _: &LabeledStmt) {}
+    fn visit_continue_stmt(&mut self, _: &ContinueStmt) {}
+
+    fn visit_arrow_expr(&mut self, _: &ArrowExpr) {}
+    fn visit_function(&mut self, _: &Function) {}
+    fn visit_constructor(&mut self, _: &Constructor) {}
+    fn visit_setter_prop(&mut self, _: &SetterProp) {}
+    fn visit_getter_prop(&mut self, _: &GetterProp) {}
+    fn visit_class_prop(&mut self, _: &ClassProp) {}
+
+    fn visit_prop_name(&mut self, prop_name: &PropName) {
+        match prop_name {
+            PropName::Computed(n) => n.visit_with(self),
+            _ => {}
+        }
+    }
+
+    fn visit_decl(&mut self, decl: &Decl) {
+        match decl {
+            Decl::Class(ref c) => {
+                c.class.super_class.visit_with(self);
+                c.class.body.visit_with(self);
+            }
+            _ => {}
+        }
+    }
+
+    fn visit_ident(&mut self, ident: &Ident) {
+        let v = self.scope.borrow().idents.get(&ident.to_id()).cloned();
+        if let Some((src, _)) = v {
+            self.top_level_idents.insert(src.clone());
+        }
+    }
+}
+
 struct CommonJs<P>
 where
     P: ImportResolver,
@@ -89,7 +159,7 @@ where
         // Used only if export * exists
         let mut exported_names: Option<Ident> = None;
 
-        // make a preliminary pass through to collect exported names ahead of time
+        // Make a preliminary pass through to collect exported names ahead of time
         for item in &items {
             match item {
                 ModuleItem::ModuleDecl(ModuleDecl::ExportNamed(NamedExport {
@@ -125,6 +195,26 @@ where
             }
         }
 
+        // Make another preliminary pass to collect all import sources and their
+        // specifiers.
+        for item in &items {
+            self.in_top_level = true;
+            match item {
+                ModuleItem::ModuleDecl(ModuleDecl::Import(import)) => {
+                    self.scope.borrow_mut().insert_import(import.clone())
+                }
+                _ => {}
+            }
+        }
+
+        // Map all top-level identifiers that match imported specifiers, and blacklist
+        // them from lazy imports.
+        let mut visitor = LazyIdentifierVisitor::new(self.scope.clone());
+        items.visit_with(&mut visitor);
+        for ident in visitor.top_level_idents {
+            self.scope.borrow_mut().lazy_blacklist.insert(ident);
+        }
+
         for item in items {
             self.in_top_level = true;
 
@@ -204,11 +294,6 @@ where
 
                     match item {
                         ModuleItem::ModuleDecl(ModuleDecl::ExportAll(export)) => {
-                            self.scope
-                                .borrow_mut()
-                                .lazy_blacklist
-                                .insert(export.src.value.clone());
-
                             let mut scope_ref_mut = self.scope.borrow_mut();
                             let scope = &mut *scope_ref_mut;
 
@@ -776,10 +861,7 @@ where
 
     fn fold_prop(&mut self, p: Prop) -> Prop {
         match p {
-            Prop::Shorthand(ident) => {
-                let top_level = self.in_top_level;
-                Scope::fold_shorthand_prop(self, top_level, ident)
-            }
+            Prop::Shorthand(ident) => Scope::fold_shorthand_prop(self, ident),
 
             _ => p.fold_children_with(self),
         }
